--[[
	Updater.lua
		Handles ranged updating + buff highlighting updating
--]]

-- Original code by Tuller. Modified by Starslayer.

Dominos_BuffTimes = LibStub("AceAddon-3.0"):NewAddon("Dominos_BuffTimes", 'AceConsole-3.0')

--[[ Dominos_BuffTimes functions ]]--

function Dominos_BuffTimes:CheckDB()
	-- populate db with default values if none are present
	if not self.db.char.ignored then
		self.db.char.ignored = {}
		self.db.char.ignored['judgement of light'] = 1
		self.db.char.ignored['mortal strike'] = 1
		
		self.db.char.yourbuffs = {}
		self.db.char.yourbuffs['renew'] = 1
		
		self.db.char.alldebuffs = {}
		self.db.char.alldebuffs['banish'] = 1
	end
end

function Dominos_BuffTimes:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("Dominos_BuffTimesDB")
	self:CheckDB()
	self:RegisterSlashCommands()
	
	self.options = {}
	self:InitOptions()
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Dominos_BuffTimes", self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Dominos_BuffTimes", "Dominos_BuffTimes")
end

function Dominos_BuffTimes:OnEnable()
    -- Called when the addon is enabled
end

function Dominos_BuffTimes:OnDisable()
    -- Called when the addon is disabled
end

--[[ Local Functions ]]--

--clear a table, returning true if there was stuff to clear
local function ClearTable(t)
	if next(t) then
		for i in pairs(t) do
			t[i] = nil
		end
		return true
	end
end

--remove any values from t that are not in toClone
--adds any values from tableToClone that are not in t
--requires that both tables be using the same key value pairs
local function CloneTable(from, to)
	local changed = false

	for i,oldVal in pairs(to) do
		local newVal = from[i]
		-- don't remove so that OmniCC pulses correctly
		if (not newVal and oldVal) then
			if (oldVal.start and oldVal.start + oldVal.duration < GetTime()) then
				to[i].duration = 0
			else
				to[i] = newVal
			end
			changed = true
		elseif (newVal and not oldVal) or (oldVal and newVal and (oldVal.start ~= newVal.start or oldVal.duration ~= newVal.duration)) then
			to[i] = newVal
			changed = true
		end
	end

	for i,newVal in pairs(from) do
		local oldVal = to[i]
		if (newVal == nil and oldVal ~= nil) or (newVal ~= nil and oldVal == nil) or (oldVal ~= nil and newVal ~= nil and (oldVal.start ~= newVal.start or oldVal.duration ~= newVal.duration)) then
			to[i] = newVal
			changed = true
		end
	end

	return changed
end

--[[
	The Code
--]]

local Updater = CreateFrame('Frame')

Updater.targetBuffs = {}
Updater.targetBuffs['target'] = {}
Updater.targetBuffs['focus'] = {}
Updater.targetDebuffs = {}
Updater.targetDebuffs['target'] = {}
Updater.targetDebuffs['focus'] = {}
Updater.playerBuffs = {}
Updater.buttons = {}

--buff and debuff caches
local newVals = {} --store new info in here


--[[ Events ]]--

--buff and debuff updating stuff
Updater:SetScript('OnEvent', function(self, event, unit)
	if event == 'PLAYER_TARGET_CHANGED' then
		self:UpdateTargetBuffs('target')
	elseif event == 'PLAYER_FOCUS_CHANGED' then
		self:UpdateTargetBuffs('focus')
	elseif event == 'UNIT_AURA' then
		if unit == 'target' or unit == 'focus' then
			self:UpdateTargetBuffs(unit)
		elseif unit == 'player' then
			self:UpdatePlayerBuffs()
		end
	elseif event == 'PLAYER_ENTERING_WORLD' then
		self:UpdateTargetBuffs('target')
		self:UpdateTargetBuffs('focus')
		self:UpdatePlayerBuffs()
	end
end)
Updater:RegisterEvent('UNIT_AURA')
Updater:RegisterEvent('PLAYER_TARGET_CHANGED')
Updater:RegisterEvent('PLAYER_FOCUS_CHANGED')
Updater:RegisterEvent('PLAYER_ENTERING_WORLD')


--on update script, handles throttled buff and debuff updating as well as range updating
Updater:SetScript('OnUpdate', function(self, elapsed)
	if self.shouldUpdateBuffs then
		self.shouldUpdateBuffs = nil

		for _,b in pairs(self.buttons) do
			if b:IsVisible() and HasAction(b.action) then
				ActionButton_UpdateState(b)
			end
		end
	end
end)
Updater.nextUpdate = 1
Updater.delay = 0.025


--[[ Update Functions ]]--

function Updater:UpdatePlayerBuffs()
	local changed = false

	ClearTable(newVals)

	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable
	local i = 1
	repeat
		name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff('player', i)
		-- we still want to highlight buffs that aren't ours
		if name then
			newVals[name] = {}
			--newVals[name].start = expirationTime - duration
			--newVals[name].duration = duration
		end
		i = i + 1
	until not name

	if CloneTable(newVals, self.playerBuffs) then
		changed = true
	end

	--something changed, trigger update buffs
	if changed then
		self.shouldUpdateBuffs = true
	end
end

function Updater:UpdateTargetBuffs(unit)
	local changed = false

	if UnitExists(unit) then
		if UnitIsFriend('player', unit) then
			changed = self:UpdateFriendlyTargetBuffs(unit)
		else
			changed = self:UpdateEnemyTargetDebuffs(unit)
		end
	else
		changed = self:ClearTargetBuffsAndDebuffs(unit)
	end

	--if change, mark for updating
	if changed then
		self.shouldUpdateBuffs = true
	end
end


function Updater:UpdateFriendlyTargetBuffs(unit)
	--friendly target, clear target debuffs
	local changed = ClearTable(self.targetDebuffs[unit])

	--add all target buffs into newVals
	ClearTable(newVals)
	local i = 1
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable
	repeat
		name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff(unit, i)
		if name then
			-- prioritise our buffs over others
			if isMine or not newVals[name] then
				newVals[name] = {}
				newVals[name].start = expirationTime - duration
				newVals[name].duration = duration
				newVals[name].isMine = isMine
			end
		end
		i = i + 1
	until not name

	--set changed to true if the target buffs table has changed
	if CloneTable(newVals, self.targetBuffs[unit]) then
		changed = true
	end

	return changed
end

function Updater:UpdateEnemyTargetDebuffs(unit)
	--friendly target, clear target debuffs
	local changed = ClearTable(self.targetBuffs[unit])

	--update debuffs on enemy targets
	ClearTable(newVals)
	local i = 1
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable
	repeat
		name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitDebuff(unit, i)
		if name then
			-- prioritise our buffs over others
			if isMine or not newVals[name] then
				newVals[name] = {}
				newVals[name].start = expirationTime - duration
				newVals[name].duration = duration
				newVals[name].isMine = isMine
			end
		end
		i = i + 1
	until not name

	--set changed to true if the target debuffs table has changed
	if CloneTable(newVals, self.targetDebuffs[unit]) then
		changed = true
	end

	return changed
end

function Updater:ClearTargetBuffsAndDebuffs(unit)
	local changed = false

	if ClearTable(self.targetBuffs[unit]) then
		changed = true
	end
	if ClearTable(self.targetDebuffs[unit]) then
		changed = true
	end

	return changed
end


--[[ Access Functions ]]--

function Updater:GetTargetBuff(unit, buff)
	return self.targetBuffs[unit][buff]
end

function Updater:GetTargetDebuff(unit, debuff)
	return self.targetDebuffs[unit][debuff]
end

function Updater:PlayerHasBuff(buff)
	return self.playerBuffs[buff]
end


--[[ Action Button Updating ]]--

local function ShouldShowTimer(spell, buff, spellBuff, cdstart, cdduration)
	if Dominos_BuffTimes:IsIgnored(spell) then return nil end
	
	if buff == 1 and Dominos_BuffTimes:IsYourBuff(spell) and not spellBuff.isMine then return nil end
	
	if buff == 0 and not Dominos_BuffTimes:IsAllDebuff(spell) and not spellBuff.isMine then return nil end
	
	if cdstart == nil or cdduration == 0 or spellBuff.start + spellBuff.duration > cdstart + cdduration then return 1 end
	
	return nil
end

local function ActionButton_UpdateBorder(self, spell)
	local buff = 1
	local showBorder = nil
	
	if spell then
		local unit = self.unit
		if UnitExists(unit) then
			local spellBuff = nil
			
			if UnitIsFriend('player', unit) then
				spellBuff = Updater:GetTargetBuff(unit, spell)
				if spellBuff then
					self:GetCheckedTexture():SetVertexColor(0, 1, 0)
					showBorder = 1
				end
			else
				buff = 0
				spellBuff = Updater:GetTargetDebuff(unit, spell)
				
				if spellBuff and spellBuff.duration == 0 then 
					return false
				end
				
				if spellBuff and spellBuff.isMine and spellBuff.duration ~= 0 then
					self:GetCheckedTexture():SetVertexColor(1, 0, 1)
					showBorder = 1
				end
			end
			
			if spellBuff and spellBuff.start ~= nil then
				local start, duration, enabled
				start, duration, enabled = GetSpellCooldown(spell)
				
				if ShouldShowTimer(spell, buff, spellBuff, start, duration) then
					if self.cooldown == nil then
						local _G = getfenv(0)
						local name = self:GetName()
						self.origcooldown = _G[name .. 'Cooldown']
						self.cooldown = CreateFrame('Cooldown', nil, self, 'CooldownFrameTemplate')
						self.cooldown:SetAllPoints(self)
					end
					self.origcooldown:Hide()
					CooldownFrame_SetTimer(self.cooldown, spellBuff.start, spellBuff.duration, 1)
				elseif self.cooldown and self.cooldown:IsShown() then
					CooldownFrame_SetTimer(self.cooldown, 0, 0, nil)
					self.origcooldown:Show()
				end
				
				return showBorder
			end
		end

		local playerBuff = Updater:PlayerHasBuff(spell)
		if playerBuff and not UnitIsFriend('player', unit) and playerBuff.duration ~= 0 then
			self:GetCheckedTexture():SetVertexColor(0, 1, 0)

			if self.cooldown ~= nil and self.cooldown:IsShown() then
				CooldownFrame_SetTimer(self.cooldown, 0, 0, nil)
				self.origcooldown:Show()
			end

			return true
		end
	end
	
	--self:GetCheckedTexture():SetVertexColor(1, 1, 1)
	
	if self.cooldown and self.cooldown:IsShown() then
		-- need to distinguish between buff dropping off and change targets
		CooldownFrame_SetTimer(self.cooldown, 0, 0, nil)
		self.origcooldown:Show()
	end
	
	return showBorder
end

local function ActionButton_CheckCooldown (self)
	if self.cooldown ~= nil then
		if self.cooldown:IsShown() then
			self.origcooldown:Hide()
		end
	end
end

hooksecurefunc('ActionButton_UpdateCooldown', ActionButton_CheckCooldown)

local function ActionButton_IsSpellInUse(self)
	local spellID = self.spellID
	if spellID then
		if self.type == 'macro' then
			return ActionButton_UpdateBorder(self, GetMacroSpell(spellID))
		end
		return ActionButton_UpdateBorder(self, spellID)
	end
end

local function ActionButton_UpdateSpell(self)
	if self.action then
		local type, arg1, arg2 = GetActionInfo(ActionButton_GetPagedID(self))
		self.unit = 'target'
		self.type = type
		if type == 'spell' then
			if arg1 and arg2 and arg1 > 0 then
				self.spellID = GetSpellName(arg1, arg2)
			else
				self.spellID = nil
			end
		elseif type == 'item' then
			self.spellID = GetItemSpell(arg1)
		else
			self.spellID = arg1
			local macroSpell = GetMacroSpell(self.spellID)
		end
		
		if type == 'macro' then
			local macroName, macroIcon, macroBody, macroLocal = GetMacroInfo(arg1)
			
			if macroName ~= nil then
				-- check if we have ' Focus' in the macro name, if so, target is 'focus' for this button
				local startPos, endPos = string.find(macroName, '%sFocus')
			
				if startPos ~= nil then
					self.unit = 'focus'
				end
			end
		end

		ActionButton_UpdateState(self)
	end
end

hooksecurefunc('ActionButton_UpdateState', function(self)
	self:SetChecked(ActionButton_IsSpellInUse(self) or self:GetChecked())
end)

hooksecurefunc('ActionButton_Update', function(self)
	ActionButton_UpdateSpell(self)
end)


--[[ Loading ]]--

--register buttons
for id = 1, 12 do
	table.insert(Updater.buttons, _G['ActionButton' .. id])
	table.insert(Updater.buttons, _G['BonusActionButton' .. id])
	table.insert(Updater.buttons, _G['MultiBarRightButton' .. id])
	table.insert(Updater.buttons, _G['MultiBarLeftButton' .. id])
	table.insert(Updater.buttons, _G['MultiBarBottomRightButton' .. id])
	table.insert(Updater.buttons, _G['MultiBarBottomLeftButton' .. id])
end

--register any stock action buttons created after this addon is loaded
hooksecurefunc('ActionButton_OnLoad', function(self)
	table.insert(Updater.buttons, self)
end)